#!/usr/bin/env python3
# -*- coding: utf-8 -*-
'''
Author: Riku Laine
Project name: Evaluating Decision-Makers over Selectively Labeled Data
Description: This script creates the figures and results used
             in the synthetic data experiments.

Parameters:
-----------
(1) figure_path : file location for saving the created figures.
(2) N_sim : Size of simulated data set.
(3) M_sim : Number of judges in simulated data,
            N_sim must be divisible by M_sim!
(4) decider_H : Which decider H to use (random, batch, independent or
                probabilistic).
(5) decider_M : Which decider M to use (random or lr (batch)).
(6) stan_code_file_name : Name of file containing the Stan model code.
(7) sigma_tau : Value of prior variance.
(8) Z_coef : Coefficient for the effect of Z.
(9) max_leniency : Upper limit of interval from which to draw leniency R.
'''

# Imports

import sys
import gc
import numpy as np
import numpy.random as npr
import pandas as pd
import matplotlib.pyplot as plt
import scipy.special as ssp
from sklearn.linear_model import LogisticRegression
from cmdstanpy import Model

plt.switch_backend('agg')

# Seed for reproducibility
npr.seed(123)

# figure storage name
figure_path = sys.argv[1]

# Size of simulated data set
N_sim = int(sys.argv[2])

# Number of judges in simulated data, N_sim must be divisible by M_sim!
M_sim = int(sys.argv[3])

# Decider H type.
decider_H = sys.argv[4]

# Decider M type
decider_M = sys.argv[5]

# Name of stan model code file
stan_code_file_name = sys.argv[6]

# Variance prior
sigma_tau = float(sys.argv[7])

# Z coefficient
Z_coef = float(sys.argv[8])

# Maximum leniency
max_leniency = float(sys.argv[9])

# Settings

plt.rcParams.update({'font.size': 16})
plt.rcParams.update({'figure.figsize': (10, 6)})

print("These results have been obtained with the following settings:")

print("Number of observations in the simulated data:", N_sim)

print("Number of judges in the simulated data:", M_sim)

print("Decider H type:", decider_H)

print("Decider M type:", decider_M)

print("Prior for the variances:", sigma_tau)

print("Z coefficient:", Z_coef)

print("Maximum leniency:", max_leniency)


### Mathematical functions

def inverseLogit(x):
    return 1.0 / (1.0 + np.exp(-1.0 * x))


def logit(x):
    return np.log(x) - np.log(1.0 - x)

def inverseCumulative(x, mu, sigma):
    '''Compute the inverse of the cumulative distribution of logit-normal
    distribution at x with parameters mu and sigma (mean and st.dev.).'''

    return inverseLogit(ssp.erfinv(2 * x - 1) * np.sqrt(2 * sigma**2) - mu)

### Convenience functions

def fitPredictiveModel(x_train, y_train, x_test, class_value, model_type=None):
    '''
    Fit a predictive model (default logistic regression) with given training
    instances and return probabilities for test instances to obtain a given
    class label.

    Arguments:
    ----------

    x_train -- x values of training instances
    y_train -- y values of training instances
    x_test -- x values of test instances
    class_value -- class label for which the probabilities are counted for.
    machine_decider_type -- type of model to be fitted.

    Returns:
    --------
    (1) Trained predictive model
    (2) Probabilities for given test inputs for given class.
    '''

    if model_type is None or model_type != "random":
        # Instantiate the model (using the default parameters)
        model = LogisticRegression(solver='lbfgs')

        # Check shape and fit the model.
        if x_train.ndim == 1:
            model = model.fit(x_train.values.reshape(-1, 1), y_train)
        else:
            model = model.fit(x_train, y_train)

        label_probs = getProbabilityForClass(x_test, model, class_value)

    elif model_type == "random":

        label_probs = npr.uniform(size=x_test.shape)

        model = lambda x: npr.uniform(size=1)

    else:
        raise ValueError("Invalid model_type:", model_type)

    return model, label_probs


def getProbabilityForClass(x, model, class_value):
    '''
    Function (wrapper) for obtaining the probability of a class given x and a
    predictive model.

    Arguments:
    -----------
    x -- individual features, an array of shape (observations, features)
    model -- a trained sklearn model. Predicts probabilities for given x.
        Should accept input of shape (observations, features)
    class_value -- the resulting class to predict (usually 0 or 1).

    Returns:
    --------
    (1) The probabilities of given class label for each x.
    '''
    if x.ndim == 1:
        # if x is vector, transform to column matrix.
        f_values = model.predict_proba(np.array(x).reshape(-1, 1))
    else:
        f_values = model.predict_proba(x)

    # Get correct column of predicted class, remove extra dimensions and return.
    return f_values[:, model.classes_ == class_value].flatten()

### Data generation modules

def bernoulliDGWithUnobservables(N_total=50000,
                                 beta_X=1.0,
                                 beta_Z=1.0):
    '''Generates data | Variables: X, Z, Y | Outcome from Bernoulli'''

    df = pd.DataFrame()

    # Sample features X and Z and noise epsilon_Y independently from standard
    # Gaussian distributions, N(0, 1).
    df = df.assign(X=npr.normal(size=N_total))
    df = df.assign(Z=npr.normal(size=N_total))
    df = df.assign(epsilon_Y=npr.normal(size=N_total))

    # Calculate P(Y=0|X=x) = 1 / (1 + exp(-X)) = inverseLogit(X)
    probabilities_Y = inverseLogit(beta_X * df.X + beta_Z * df.Z +
                                   np.sqrt(0.1) * df.epsilon_Y)

    df = df.assign(probabilities_Y=probabilities_Y)

    # Draw Y from Bernoulli distribution
    results = npr.binomial(n=1, p=1 - df.probabilities_Y, size=N_total)

    df = df.assign(result_Y=results)

    return df


### Decider H modules

def humanBatchDeciderLakkaraju(df,
                               featureX_col,
                               featureZ_col=None,
                               nJudges_M=100,
                               beta_X=1,
                               beta_Z=1,
                               leniency_upper_limit=0.9,
                               add_epsilon=True):
    '''Decider module | Non-independent batch decisions.'''
    # Decider as specified by Lakkaraju et al.

    # Assert that every judge will have the same number of subjects.
    assert df.shape[0] % nJudges_M == 0, "Can't assign subjets evenly!"

    # Compute the number of subjects allocated for each judge.
    nSubjects_N = int(df.shape[0] / nJudges_M)

    # Assign judge IDs as running numbering from 0 to nJudges_M - 1
    df = df.assign(judgeID_J=np.repeat(range(0, nJudges_M), nSubjects_N))

    # Sample acceptance rates uniformly from a closed interval
    # from 0.1 to 0.9 and round to tenth decimal place.
    acceptance_rates = npr.uniform(.1, leniency_upper_limit, nJudges_M)
    acceptance_rates = np.round(acceptance_rates, 10)

    # Replicate the rates so they can be attached to the corresponding judge ID.
    # True leniency is observed leniency.
    df = df.assign(acceptanceRate_R=np.repeat(acceptance_rates, nSubjects_N))
    df = df.assign(true_leniency=np.repeat(acceptance_rates, nSubjects_N))

    # Add noise
    if add_epsilon:
        epsilon_T = np.sqrt(0.1) * npr.normal(size=df.shape[0])
    else:
        epsilon_T = 0

    # Compute probability for jail decision
    if featureZ_col is None:
        probabilities_T = inverseLogit(beta_X * df[featureX_col] + epsilon_T)
    else:
        probabilities_T = inverseLogit(beta_X * df[featureX_col] +
                                       beta_Z * df[featureZ_col] + epsilon_T)

    df = df.assign(probabilities_T=probabilities_T)

    # Sort by judges then probabilities in decreasing order
    # Most dangerous for each judge are at the top.
    df.sort_values(by=["judgeID_J", "probabilities_T"],
                   ascending=False,
                   inplace=True)

    # Iterate over the data. Subject will be given a negative decision
    # if they are in the top (1-r)*100% of the individuals the judge will judge.
    # I.e. if their within-judge-index is under 1 - acceptance threshold times
    # the number of subjects assigned to each judge they will receive a
    # negative decision.
    df.reset_index(drop=True, inplace=True)

    df['decision_T'] = np.where((df.index.values % nSubjects_N) <
                                ((1 - df['acceptanceRate_R']) * nSubjects_N),
                                0, 1)

    df_labeled = df.copy()

    # Hide unobserved
    df_labeled.loc[df.decision_T == 0, 'result_Y'] = np.nan

    return df_labeled, df


def probabilisticDecider(df,
                         featureX_col,
                         featureZ_col=None,
                         nJudges_M=100,
                         beta_X=1,
                         beta_Z=1,
                         add_epsilon=True,
                         leniency_upper_limit=0.9):
    '''Use X and Z to make a decision with probability
    P(T=0|X, Z, R)=inverseLogit(inverseLogit(R)+beta_X*X+beta_Z*Z).'''

    # Assert that every judge will have the same number of subjects.
    assert df.shape[0] % nJudges_M == 0, "Can't assign subjets evenly!"

    # Compute the number of subjects allocated for each judge.
    nSubjects_N = int(df.shape[0] / nJudges_M)

    # Assign judge IDs as running numbering from 0 to nJudges_M - 1
    df = df.assign(judgeID_J=np.repeat(range(0, nJudges_M), nSubjects_N))

    # Sample acceptance rates uniformly from a closed interval
    # from 0.1 to 0.9 and round to tenth decimal place.
    acceptance_rates = npr.uniform(.1, leniency_upper_limit, nJudges_M)
    acceptance_rates = np.round(acceptance_rates, 10)

    # Replicate the rates so they can be attached to the corresponding judge ID.
    df = df.assign(true_leniency=np.repeat(acceptance_rates, nSubjects_N))

    if add_epsilon:
        epsilon_T = np.sqrt(0.1) * npr.normal(size=df.shape[0])
    else:
        epsilon_T = 0

    if featureZ_col is None:
        probabilities_T = inverseLogit(logit(df.true_leniency) +
                                       beta_X * df[featureX_col] + epsilon_T)
    else:
        probabilities_T = inverseLogit(logit(df.true_leniency) +
                                       beta_X * df[featureX_col] +
                                       beta_Z * df[featureZ_col] + epsilon_T)

    df = df.assign(probabilities_T=probabilities_T)

    # Draw T from Bernoulli distribution
    decisions = npr.binomial(n=1, p=1 - df.probabilities_T, size=df.shape[0])

    df = df.assign(decision_T=decisions)

    # Calculate the observed acceptance rates. NB: observed acceptance rate !=
    # true leniency.
    acceptance_rates = df.groupby('judgeID_J').mean().decision_T.values

    # Replicate the rates so they can be attached to the corresponding judge ID.
    df = df.assign(acceptanceRate_R=np.repeat(acceptance_rates, nSubjects_N))

    df_labeled = df.copy()

    df_labeled.loc[df.decision_T == 0, 'result_Y'] = np.nan

    return df_labeled, df


def humanIndependentDecider(df, featureX_col, featureZ_col=None, nJudges_M=100,
                       beta_X=1, beta_Z=1, add_epsilon=True,
                       leniency_upper_limit=0.9):

    '''Assign decisions by the value of inverse cumulative distribution function
    of the logit-normal distribution at leniency r.'''

    # Assert that every judge will have the same number of subjects.
    assert df.shape[0] % nJudges_M == 0, "Can't assign subjets evenly!"

    # Compute the number of subjects allocated for each judge.
    nSubjects_N = int(df.shape[0] / nJudges_M)

    # Assign judge IDs as running numbering from 0 to nJudges_M - 1
    df = df.assign(judgeID_J=np.repeat(range(0, nJudges_M), nSubjects_N))

    # Sample acceptance rates uniformly from a closed interval
    # from 0.1 to 0.9 and round to tenth decimal place.
    acceptance_rates = npr.uniform(.1, leniency_upper_limit, nJudges_M)
    acceptance_rates = np.round(acceptance_rates, 10)

    # Replicate the rates so they can be attached to the corresponding judge ID.
    df = df.assign(true_leniency=np.repeat(acceptance_rates, nSubjects_N))

    if add_epsilon:
        epsilon_T = np.sqrt(0.1) * npr.normal(size=df.shape[0])
    else:
        epsilon_T = 0

    if featureZ_col is None:
        probabilities_T = inverseLogit(beta_X * df[featureX_col] + epsilon_T)

        # Compute the bounds straight from the inverse cumulative.
        # Assuming X is N(0, 1) so Var(bX*X)=bX**2*Var(X)=bX**2.
        df = df.assign(bounds=inverseCumulative(
            x=df.true_leniency, mu=0, sigma=np.sqrt(beta_X**2)))
    else:
        probabilities_T = inverseLogit(beta_X * df[featureX_col] +
                                       beta_Z * df[featureZ_col] + epsilon_T)

        # Compute the bounds straight from the inverse cumulative.
        # Assuming X and Z are i.i.d standard Gaussians with variance 1.
        # Thus Var(bx*X+bZ*Z)= bX**2*Var(X)+bZ**2*Var(Z).
        df = df.assign(bounds=inverseCumulative(
            x=df.true_leniency, mu=0, sigma=np.sqrt(beta_X**2 + beta_Z**2)))

    df = df.assign(probabilities_T=probabilities_T)

    # Assign negative decision if the predicted probability (probabilities_T) is
    # over the judge's threshold (bounds).
    df = df.assign(decision_T=np.where(df.probabilities_T >= df.bounds, 0, 1))

    # Calculate the observed acceptance rates.
    acceptance_rates = df.groupby('judgeID_J').mean().decision_T.values

    # Replicate the rates so they can be attached to the corresponding judge ID.
    df = df.assign(acceptanceRate_R = np.repeat(acceptance_rates, nSubjects_N))

    df_labeled = df.copy()

    df_labeled.loc[df.decision_T == 0, 'result_Y'] = np.nan

    return df_labeled, df


def humanRandomDecider(df, nJudges_M=100, use_acceptance_rates=False,
                  leniency_upper_limit=0.9):
    '''Doesn't use any information about X and Z to make decisions.

    If use_acceptance_rates is False (default) then all decisions are positive
    with probability 0.5. If True, probabilities will be sampled from
    U(0.1, 0.9) and rounded to tenth decimal place.'''

    # Assert that every judge will have the same number of subjects.
    assert df.shape[0] % nJudges_M == 0, "Can't assign subjets evenly!"

    # Compute the number of subjects allocated for each judge.
    nSubjects_N = int(df.shape[0] / nJudges_M)

    # Assign judge IDs as running numbering from 0 to nJudges_M - 1
    df = df.assign(judgeID_J=np.repeat(range(0, nJudges_M), nSubjects_N))

    if use_acceptance_rates:
        # Sample acceptance rates uniformly from a closed interval
        # from 0.1 to 0.9 and round to tenth decimal place.
        acceptance_rates = npr.uniform(.1, leniency_upper_limit, nJudges_M)
        acceptance_rates = np.round(acceptance_rates, 10)
    else:
        # No real leniency here -> set to 0.5.
        acceptance_rates = np.ones(nJudges_M) * 0.5

    # Replicate the rates so they can be attached to the corresponding judge ID.
    df = df.assign(true_leniency=np.repeat(acceptance_rates, nSubjects_N))

    df = df.assign(
        decision_T=npr.binomial(n=1, p=df.true_leniency, size=df.shape[0]))

    # Calculate the observed acceptance rates.
    acceptance_rates = df.groupby('judgeID_J').mean().decision_T.values

    # Replicate the rates so they can be attached to the corresponding judge ID.
    df = df.assign(acceptanceRate_R = np.repeat(acceptance_rates, nSubjects_N))

    df_labeled = df.copy()

    df_labeled.loc[df.decision_T == 0, 'result_Y'] = np.nan

    return df_labeled, df


### Decider M modules

def machineBatchDecider(df, resultY_col, r):
    # Sort values by predicted probability for Y=0.
    df.sort_values(by='B_prob_0_model', inplace=True, ascending=True)

    # Set first r % to be released.
    to_release = int(round(df.shape[0] * r))

    # Now if Y[i] = 0, then failures[i] = 1. This now applies also for
    # CFBI: if Y[i] = 0.3, then failures[i] = 0.7.
    failures = 1 - df[resultY_col][0:to_release]

    FR_estimate = np.sum(failures) / df.shape[0]

    return FR_estimate

def machineIndependentDecider(df, resultY_col, r, referenceTableForIndependentM):
    # Solve the boundary for negative decision. Reference table declares
    # probabilities for Y = 0.
    bound = np.quantile(referenceTableForIndependentM, r)

    # If predicted probability is under bound, release.
    to_release = np.where(df['B_prob_0_model'] < bound, True, False)

    print("Desired AR: " + str(r) + ", obs. AR: " + str(np.mean(to_release)))

    # Now if Y[i] = 0, then failures[i] = 1. This now applies also for
    # CFBI: if Y[i] = 0.3, then failures[i] = 0.7.
    failures = 1 - df.loc[to_release, resultY_col]

    FR_estimate = np.sum(failures) / df.shape[0]

    return FR_estimate

def machineRandomDecider(df, resultY_col, r):
    # Give decisions at random with probability r.
    decisions = npr.binomial(n=1, p=r, size=df.shape[0])

    # If predicted probability is under bound, release.
    to_release = np.where(decisions == 1, True, False)

    # Now if Y[i] = 0, then failures[i] = 1. This now applies also for
    # CFBI: if Y[i] = 0.3, then failures[i] = 0.7.
    failures = 1 - df.loc[to_release, resultY_col]

    FR_estimate = np.sum(failures) / df.shape[0]

    return FR_estimate

### Evaluator modules

# ### Contraction algorithm ###
#
# Below is an implementation of Lakkaraju's team's algorithm presented in
# their paper (https://helka.finna.fi/PrimoRecord/pci.acm3098066). Relevant
# parameters passed to the function are presented in the description.

def contraction(df, judgeIDJ_col, decisionT_col, resultY_col, modelProbS_col,
                accRateR_col, r):
    '''
    This is an implementation of the algorithm presented by Lakkaraju
    et al. in their paper "The Selective Labels Problem: Evaluating
    Algorithmic Predictions in the Presence of Unobservables" (2017).

    Arguments:
    ----------
    df -- The (Pandas) data frame containing the data, judge decisions,
        judge IDs, results and probability scores.
    judgeIDJ_col -- String, the name of the column containing the judges' IDs
        in df.
    decisionT_col -- String, the name of the column containing the judges' decisions
    resultY_col -- String, the name of the column containing the realization
    modelProbS_col -- String, the name of the column containing the probability
        scores from the black-box model B.
    accRateR_col -- String, the name of the column containing the judges'
        acceptance rates
    r -- Float between 0 and 1, the given acceptance rate.

    Returns:
    --------
    (1) The estimated failure rate at acceptance rate r.
    '''
    # Get ID of the most lenient judge.
    most_lenient_ID_q = df[judgeIDJ_col].loc[df[accRateR_col].idxmax()]

    # Subset. "D_q is the set of all observations judged by q."
    D_q = df[df[judgeIDJ_col] == most_lenient_ID_q].copy()

    ##### Agreement rate computation begins #######

    Q_leniency = max(df[accRateR_col])

    if Q_leniency < r:
        return np.nan, np.nan

    # Sort D_q
    # "Observations deemed as high risk by B are at the top of this list"
    D_sort_q = D_q.sort_values(by=modelProbS_col, ascending=False)

    # The agreement rate is now computed as the percentage of all the positive
    # decisions given by q in the 1-r % of the most dangerous subjects.
    all_negatives = np.sum(D_sort_q[decisionT_col] == 0)

    n_most_dangerous = int(round(D_q.shape[0] * (1-Q_leniency)))

    agreed_decisions = np.sum(D_sort_q[decisionT_col][0:n_most_dangerous] == 0)

    print("The agreement rate of contraction is:", agreed_decisions/all_negatives)

    ##### Agreement rate computation ends #######

    # All observations of R_q have observed outcome labels.
    # "R_q is the set of observations in D_q with observed outcome labels."
    R_q = D_q[D_q[decisionT_col] == 1].copy()

    # Sort observations in R_q in descending order of confidence scores S and
    # assign to R_sort_q.
    # "Observations deemed as high risk by B are at the top of this list"
    R_sort_q = R_q.sort_values(by=modelProbS_col, ascending=False)

    number_to_remove = int(
        round((1.0 - r) * D_q.shape[0] - (D_q.shape[0] - R_q.shape[0])))

    # "R_B is the list of observations assigned to t = 1 by B"
    R_B = R_sort_q[number_to_remove:R_sort_q.shape[0]]

    return np.sum(R_B[resultY_col] == 0) / D_q.shape[0], D_q.shape[0]


# Other evaluators

def trueEvaluationEvaluator(df, resultY_col, r, machineDeciderType,
                            referenceTableForIndependentM=None):

    if machineDeciderType == "batch":
        FR_estimate = machineBatchDecider(df, resultY_col, r)

    elif machineDeciderType == "independent":
        FR_estimate = machineIndependentDecider(df, resultY_col, r,
                                                referenceTableForIndependentM)

    elif machineDeciderType == "random":
        FR_estimate = machineRandomDecider(df, resultY_col, r)

    return FR_estimate, df.shape[0]


def labeledOutcomesEvaluator(df, decisionT_col, resultY_col, r, machineDeciderType,
                             referenceTableForIndependentM=None):

    df_observed = df.loc[df[decisionT_col] == 1, :]

    if machineDeciderType == "batch":
        FR_estimate = machineBatchDecider(df_observed, resultY_col, r)

    elif machineDeciderType == "independent":
        FR_estimate = machineIndependentDecider(df_observed, resultY_col, r,
                                                referenceTableForIndependentM)

    elif machineDeciderType == "random":
        FR_estimate = machineRandomDecider(df_observed, resultY_col, r)

    return FR_estimate, df.shape[0]


def imputationEvaluator(df, decisionT_col, resultY_col, featureX_col, r,
                        machineDeciderType, referenceTableForIndependentM=None):
    '''
    Imputation evaluator:

    In the labeled test data (i.e. data where outcomes Y have been set to 1 if
    decision T has been 0), impute the missing outcomes Y based on the observed
    variables X with T=1.
    '''

    df_imputed = df.copy()

    __, probabilities = fitPredictiveModel(df[featureX_col][df[decisionT_col]==1],
                                           df[resultY_col][df[decisionT_col]==1],
                                           df[featureX_col], 1)

    # Impute outcome from model if T == 0.
    df_imputed[resultY_col] = np.where(df[decisionT_col]==0,
                                       np.round(probabilities).astype(int),
                                       df[resultY_col])

    # Then we can employ the true evaluation evaluator.

    return trueEvaluationEvaluator(df_imputed, resultY_col, r, machineDeciderType,
                                   referenceTableForIndependentM)

### Other functions

def drawDiagnostics(title, save_name, f_rates, titles):
    '''
    Draw boxplot figures. One panel for each method.
    '''

    cols = 2
    rows = np.ceil(len(f_rates) / cols)

    plt.figure(figsize=(16, 4.5*rows+1))

    ax = plt.subplot(rows, cols, 1)
    x_ax = np.arange(1, 9, 1) / 10

    plt.boxplot(f_rates[0], labels=x_ax)

    plt.title(titles[0])
    plt.xlabel('Acceptance rate')
    plt.ylabel('Failure rate')
    plt.grid()

    for i in range(len(f_rates)):
        plt.subplot(rows, cols, i + 1, sharey=ax)

        plt.boxplot(f_rates[i], labels=x_ax)

        plt.title(titles[i])
        plt.xlabel('Acceptance rate')
        plt.ylabel('Failure rate')
        plt.grid()

    plt.tight_layout()
    plt.subplots_adjust(top=0.85)
    plt.suptitle(title, y=0.96, weight='bold')

    plt.savefig(save_name + '_diagnostic_plot')

    plt.show()


def perfComp(dgModule, deciderModule, title, save_name, machineDeciderType,
             nIter=10, true_Z_coefficient=1):
    '''
    Performance comparison, a.k.a the main figure.

    Parameters:
    -----------
    dgModule: Module creating the data
    deciderModule: Module assigning the decisions
    title: Title for the diagnostic figure
    save_name: name for saving
    machine_decider_type: Type of machine decider ("random", "batch" or
                          "independent").
    nIter: Number of train-test splits to be performed.
    true_Z_coefficient: Coefficient in decision-making and data generation.
    '''

    failure_rates = np.zeros((8, 5))
    failure_stds = np.zeros((8, 5))

    f_rate_true = np.zeros((nIter, 8))
    f_rate_label = np.zeros((nIter, 8))
    f_rate_imp = np.zeros((nIter, 8))
    f_rate_cont = np.zeros((nIter, 8))
    f_rate_cf = np.zeros((nIter, 8))

    for i in range(nIter):

        # Create data
        df = dgModule()

        # Assign decicions
        df_labeled, df_unlabeled = deciderModule(df)

        # Split data using judge IDs.

        # Get number of judges
        id_max = np.max(df_labeled.judgeID_J) + 1

        # Choose random judges by id
        selected = npr.choice(np.arange(id_max), int(id_max/2), False)

        test_labeled = df_labeled[df_labeled.judgeID_J.isin(selected)]
        train = df_labeled[~df_labeled.judgeID_J.isin(selected)]

        # Renew IDs
        test_labeled.loc[:, 'judgeID_J'] = np.repeat(np.arange(int(id_max/2)),
                        int(test_labeled.shape[0]/(id_max/2)))

        # Assign same observations to unlabeled data
        test_unlabeled = df_unlabeled.iloc[test_labeled.index.values]

        # Train the predictive model with observations in the training data.
        # Use only logistic regression model.

        B_model, predictions = fitPredictiveModel(
            train.loc[train['decision_T'] == 1, 'X'],
            train.loc[train['decision_T'] == 1, 'result_Y'],
            test_labeled['X'], 0, model_type=machineDeciderType)

        # Build reference table for independent M from predictions of the
        # model built above on the training data. the reference table contains
        # an approximate distribution of the values from the predictive model.
        # Values represent probability for Y=0.

        if machineDeciderType == "random":
            referenceTableForIndependentM = npr.uniform(size=train.shape[0])
        else:
            referenceTableForIndependentM = getProbabilityForClass(x=train['X'],
                                                                   model=B_model,
                                                                   class_value=0)

        # Attach predictions to data
        test_labeled = test_labeled.assign(B_prob_0_model=predictions)
        test_unlabeled = test_unlabeled.assign(B_prob_0_model=predictions)

        # Sort the subjects based the models prediction.
        test_labeled.sort_values(by='B_prob_0_model', inplace=True, ascending=True)

        ###### CFBI evaluation starts ######

        # Find observed values
        observed = test_labeled['decision_T'] == 1

        # Assign data to the model

        ###############################
        # Change this assignment if you want to use predictions as input to the
        # model.
        ###############################
        dat = dict(D=int(1),
                   N_obs=int(np.sum(observed)),
                   N_cens=int(np.sum(~observed)),
                   sigma_tau=sigma_tau,
                   M=int(len(set(test_labeled['judgeID_J']))),
                   jj_obs=test_labeled.loc[observed, 'judgeID_J']+1,
                   jj_cens=test_labeled.loc[~observed, 'judgeID_J']+1,
                   dec_obs=test_labeled.loc[observed, 'decision_T'],
                   dec_cens=test_labeled.loc[~observed, 'decision_T'],
                   X_obs=test_labeled.loc[observed, 'X'].values.reshape(-1, 1),
                   X_cens=test_labeled.loc[~observed, 'X'].values.reshape(-1, 1),
                   y_obs=test_labeled.loc[observed, 'result_Y'].astype(int))

        # Do the sampling - Change this if you wish to use other methods.
        # See also Stan website https://mc-stan.org for information on
        # the prerequisites for reproducing the results.
        # CMDstan sampling

        fit = sm.sample(data=dat, chains=4, seed=123, warmup_iters=1500,
                        sampling_iters=4000, adapt_delta=0.95)

        # Extract predictions
        fit_colnames = fit.column_names

        pred_cols = [s for s in fit_colnames if 'y_est' in s]

        pars = fit.get_drawset(params=pred_cols)

        plt.close()
        plt.close('all')
        gc.collect()
        gc.collect()

        print(fit.diagnose(), file=open(save_name + '_stan_fit_diagnostics_' +
              str(i) + '.txt', 'w'))

        # Format imputation matrix, each row is a sample of the posterior
        # columns are observations.

        y_imp = np.ones((pars.shape[0], test_labeled.shape[0]))

        # Impute the unobserved with the estimated
        y_imp[:, ~observed] = pars

        # Remove the variables to conserve memory
        pars = None
        fit = None

        del pars
        del fit

        gc.collect()
        gc.collect()

        # Impute the observed with observed outcomes.
        y_imp[:, observed] = test_labeled.loc[observed, 'result_Y']

        for r in range(1, 9):

            # Compute Y estimate as the mean of each column (i.e. expectation
            # of posterior for subjects with T=0 and the result it self for
            # subjects with T=1).
            Y_estimates = np.mean(y_imp, axis=1)

            test_labeled = test_labeled.assign(Y_estimates=Y_estimates)

            # Now employ true evaluation over the imputed dataset, but use
            # the estimated Y as the results.

            f_rate_cf[i, r - 1] = trueEvaluationEvaluator(df=test_labeled,
                                                          resultY_col='Y_estimates',
                                                          r=r / 10,
                                                          machineDeciderType=machineDeciderType,
                                                          referenceTableForIndependentM=referenceTableForIndependentM)

            ##### CFBI evaluation ends #####

            # Logistic regression imputation evaluator

            f_rate_imp[i, r - 1], N = imputationEvaluator(df=test_labeled,
                                                          decisionT_col='decision_T',
                                                          resultY_col='result_Y',
                                                          featureX_col='X',
                                                          r=r / 10,
                                                          machineDeciderType=machineDeciderType,
                                                          referenceTableForIndependentM=referenceTableForIndependentM)

            # True evaluation

            f_rate_true[i, r - 1], N = trueEvaluationEvaluator(df=test_unlabeled,
                                                               resultY_col='result_Y',
                                                               r=r / 10,
                                                               machineDeciderType=machineDeciderType,
                                                               referenceTableForIndependentM=referenceTableForIndependentM)

            # Labeled outcomes only

            f_rate_label[i, r - 1], N = labeledOutcomesEvaluator(df=test_labeled,
                                                                 decisionT_col='decision_T',
                                                                 resultY_col='result_Y',
                                                                 r=r / 10,
                                                                 machineDeciderType=machineDeciderType,
                                                                 referenceTableForIndependentM=referenceTableForIndependentM)

            # Contraction

            f_rate_cont[i, r - 1], N = contraction(test_labeled, 'judgeID_J',
                       'decision_T', 'result_Y', 'B_prob_0_model',
                       'acceptanceRate_R', r / 10)

    failure_rates[:, 0] = np.nanmean(f_rate_true, axis=0)
    failure_rates[:, 1] = np.nanmean(f_rate_label, axis=0)
    failure_rates[:, 2] = np.nanmean(f_rate_imp, axis=0)
    failure_rates[:, 3] = np.nanmean(f_rate_cont, axis=0)
    failure_rates[:, 4] = np.nanmean(f_rate_cf, axis=0)

    # Compute sample std
    failure_stds[:, 0] = np.nanstd(f_rate_true, axis=0, ddof=1)
    failure_stds[:, 1] = np.nanstd(f_rate_label, axis=0, ddof=1)
    failure_stds[:, 2] = np.nanstd(f_rate_imp, axis=0, ddof=1)
    failure_stds[:, 3] = np.nanstd(f_rate_cont, axis=0, ddof=1)
    failure_stds[:, 4] = np.nanstd(f_rate_cf, axis=0, ddof=1)

    x_ax = np.arange(0.1, 0.9, 0.1)

    labels = ['True Evaluation', 'Labeled outcomes', 'Logistic regression',
              'Contraction', 'CFBI']

    colours = ['g', 'magenta', 'darkmagenta', 'b', 'r']

    # General plot: Failure rate vs. Acceptance rate
    for i in range(failure_rates.shape[1]):
        plt.errorbar(x_ax,
                     failure_rates[:, i],
                     label=labels[i],
                     c=colours[i],
                     yerr=failure_stds[:, i])

    plt.xlabel('Acceptance rate')
    plt.ylabel('Failure rate')
    plt.legend(title="$\\bf{Evaluators}$")
    plt.grid()

    plt.savefig(save_name + '_all')

    plt.show()
    plt.close('all')
    gc.collect()
    gc.collect()

    # Plot error of the methods (contraction and counterfactual)
    # w.r.t True evaluation
    for i in range(3, failure_rates.shape[1]):

        error = np.abs(failure_rates[:, 0] - failure_rates[:, i])

        plt.errorbar(x_ax,
                     error,
                     label=labels[i],
                     c=colours[i],
                     yerr=failure_stds[:, i])

    plt.xlabel('Acceptance rate')
    plt.ylabel('Error w.r.t. True evaluation')
    plt.legend(title="$\\bf{Evaluators}$")
    plt.grid()

    plt.savefig(save_name + '_all_err')

    plt.show()

    plt.close('all')
    gc.collect()
    gc.collect()

    print("\nSingle runs:")
    print(f_rate_true, f_rate_label, f_rate_imp, f_rate_cont, f_rate_cf)

    # Save arrays
    np.save(save_name + "_true_FRs", f_rate_true)
    np.save(save_name + "_labeled_FRs", f_rate_label)
    np.save(save_name + "_imputed_FRs", f_rate_imp)
    np.save(save_name + "_contraction_FRs", f_rate_cont)
    np.save(save_name + "_counterfactuals_FRs", f_rate_cf)

    print("\nFailure rates:")
    print(np.array2string(failure_rates, formatter={'float_kind':lambda x: "%.5f" % x}))

    print("\nStandard deviations:")
    print(np.array2string(failure_stds, formatter={'float_kind':lambda x: "%.5f" % x}))

    print("\nMean absolute errors:")
    for i in range(1, failure_rates.shape[1]):
        print(
            labels[i].ljust(len(max(labels, key=len))),
            np.round(
                np.mean(np.abs(failure_rates[:, 0] - failure_rates[:, i])), 5))

    drawDiagnostics(title=title,
                save_name=save_name,
                f_rates=[f_rate_true, f_rate_label, f_rate_imp, f_rate_cont,
                         f_rate_cf],
                titles=labels)

# Compile stan model
sm = Model(stan_file=stan_code_file_name)
sm.compile()


dg = lambda: bernoulliDGWithUnobservables(N_total=N_sim, beta_Z=Z_coef)

if decider_H == "random":
    decider = lambda x: humanRandomDecider(x, nJudges_M=M_sim,
                                      use_acceptance_rates=True,
                                      leniency_upper_limit=max_leniency)
elif decider_H == "batch":
    decider = lambda x: humanBatchDeciderLakkaraju(x, 'X', 'Z', beta_X=1,
                                                   beta_Z=Z_coef, nJudges_M=M_sim,
                                                   leniency_upper_limit=max_leniency,
                                                   add_epsilon=False)

elif decider_H == "independent":
    decider = lambda x: humanIndependentDecider(x, 'X', 'Z', beta_X=1, beta_Z=Z_coef,
                                        nJudges_M=M_sim, add_epsilon=True,
                                        leniency_upper_limit=max_leniency)
elif decider_H == "probabilistic":
    decider = lambda x: probabilisticDecider(x, 'X', 'Z', beta_X=1,
                                             beta_Z=Z_coef, nJudges_M=M_sim,
                                             leniency_upper_limit=max_leniency)
else:
    raise ValueError("Invalid decider_H type.")

# Create log file name.
filename = figure_path + "_deciderH_" + decider_H + "_deciderM_" + decider_M

filename += "_maxR_" + str(max_leniency).replace(".", "_") + "coefZ"

filename += str(Z_coef).replace(".", "_")

### Analysis

perfComp(
    dg, decider,
    "Fluctuation of failure rate estimates across iterations",
    save_name=filename,
    machineDeciderType=decider_M,
    true_Z_coefficient=Z_coef
)
